;-------------------------------------------------------------------------
;				Program Flash / EEPROM
;-------------------------------------------------------------------------
;Input: addrh:addrl = Address
;
Verify:
	sbr	flags, 1<<f_verify
Program:
	movw	zh:zl, addrh:addrl
	andi	zh, ~high( xpage-1 )		;clear lower bits in xpage
	andi	zl, ~low( xpage-1 )		;clear lower bits in xpage
	ldi	yh, high( xpage_buff )		;buffer start
	ldi	yl, low( xpage_buff )

_prg1:	out	EEARL, zl			;1
	out	EEARH, zh			;1
	sbi	EECR, EERE			;6
	lpm	r0, Z+				;3  read Flash
	sbrc	flags, f_eeprom			;1  or
	in	r0, EEDR			;1  read EEPROM
	st	Y+, r0				;2
	cpi	yh, high( xpage_buff_end )	;1
	brne	_prg1				;2  512 * 18 = 9216 cycle

	movw	zh:zl, addrh:addrl
	andi	zh, high( xpage - 1 )
	subi	zh, high( -xpage_buff )		;start address in buffer

_prg2:	rcall	getchar				;getchar use Y !
	cpi	a0, 0xa5
	brne	_prg4
_prg3:	rcall	getchar
	cpi	a0, 0xa5
	breq	_prg3
	inc	a0
	breq	progxpage			;A0,FF = program end
	ldi	a0, 0xa5

_prg4:	ld	r0, Z				;get old byte
	st	Z+, a0				;store new byte
	cpse	r0, a0				;compare
	sbr	flags, 1<<f_write		;not equal -> write needed
	sbrs	flags, f_write
	adiw	addrh:addrl, 1			;lowest address not equal
	cpi	zh, high( xpage_buff_end )
	brne	_prg2
	rcall	progxpage
	sbrc	flags, f_error
	ret					;program / verify error
	ldi	a0, key_continue
	rcall	putchar				;o.k.
	rjmp	program				;next xpage
;-------------------------------------------------------------------------
ProgXpage:					;programm all 8 pages
	sbrs	flags, f_write
	ret					;no write needed
	cpi	addrh, high( Base * 2 )
	sbrs	flags, f_verify			;verify error
	brcs	_prx1				;or address out of range
	sbr	flags, 1<<f_error
	rjmp	hexout3

_prx1:	ldi	yh, high( xpage_buff )		;buffer start
	ldi	yl, low( xpage_buff )

	andi	addrh, ~high( xpage-1 )		;clear lower bits in xpage
	andi	addrl, ~low( xpage-1 )		;clear lower bits in xpage

_prx2:	movw	zh:zl, addrh:addrl		;page start
	ldi	a1, low( PAGESIZE )		;max 256 words (512 bytes)

_prx3:	ld	r0, Y+
	sbrs	flags, f_eeprom
	rjmp	_prx5

	out	EEARH, addrh
	out	EEARL, addrl
	adiw	addrh:addrl, 1
	sbi	EECR, EERE			;read EEPROM
	in	r1, EEDR
	cp	r0, r1
	breq	_prx6				;skip, if equal

	out	EEDR, r0
	cli
	sbi	EECR, EEMWE
	sbi	EECR, EEWE			;write EEPROM
	sei
_prx4:	sbic	EECR, EEWE			;wait until write done
	rjmp	_prx4
	rjmp	_prx6

_prx5:	ld	r1, Y+
	ldi	a0, 1<<SPMEN			;fill page buffer
	rcall	do_spm
	adiw	zh:zl, 2
	dec	a1
	brne	_prx3

	movw	zh:zl, addrh:addrl		;page start
	ldi	a0, 1<<SPMEN^1<<PGERS		;erase page
	rcall	do_spm
	ldi	a0, 1<<SPMEN^1<<PGWRT		;write page
	rcall	do_spm
	ldi	a0, 1<<SPMEN^1<<RWWSRE		;reenable RWW
	rcall	do_spm

	subi	addrl, low( -PAGESIZEB )	;next page
	sbci	addrh, high( -PAGESIZEB )

_prx6:	cpi	yh, high( xpage_buff_end )
	brne	_prx2
	cbr	flags, 1<<f_write		;write done
	ret
;-------------------------------------------------------------------------
do_spm:	cli
	out	SPMCR, a0
	spm				;do SPM command
	sei
_dos1:	in	a0, SPMCR
	sbrc	a0, SPMEN		;wait until done
	rjmp	_dos1
	ret
;-------------------------------------------------------------------------
